function [x,xi,eta] = tpls(c,r,b,xi,eta,piv)
%TPLS	least-squares solution of a Toeplitz overdetermined linear system.
%   x = TPLS(c,r,b) computes the solution, where (c,r) are the
%   first column/row of the Toeplitz matrix and b is the RHS
%   (which is unstructured and may contain multiple columns). 
%   [x,xi,eta] = TPLS(...) returns also the optimal parameters xi
%   and eta.
%   x = TPLS(c,r,b,xi,eta,piv) allows to pass the values of the
%   parameters xi and eta, and the choice for pivoting: 1
%   (partial pivoting, default) or 0 (no pivoting). 
%   If xi and eta are missing or empty, the optimal values are used.
%
%   The algorithm is described in the paper
%      G. Rodriguez.
%      Fast solution of Toeplitz- and Cauchy-like least squares problems.
%      SIAM J. Matr. Anal. Appl., 28(3):724-748, 2006.

% 	Giuseppe Rodriguez,  13-10-07
% 	Email: rodriguez@unica.it

if nargin < 6
	piv = 1;
end

c = c(:);
r = r(:);
m = size(c,1);
n = size(r,1);
kb = size(b,2);

if (nargin<5) || isempty(xi)
	phi = xieta(m,n);
	xi = 1;
	eta = exp(sqrt(-1)*pi*phi);
end

if c(1)~=r(1)
	r(1) = c(1);
	warning('Column wins diagonal conflict.')
end

if isreal([c;r;b(:)])
	reale = 1;
else
	reale = 0;
end

[GA,HA] = t2tl(c,r,xi,eta);

[tC,sC,GC,HC] = tl2cl(xi,eta,GA,HA);
GCs = spdiags(sC,0,n,n) * HC;
HCs = spdiags(conj(tC),0,m,m) * GC;

GMC = zeros(m+2*n,4);
HMC = zeros(2*m+n,4);

GMC(1:m,1:2) = GC;
GMC(m+1:m+n,3:4) = GCs;

HMC(1:m,3:4) = HCs;
HMC(m+1:m+n,1:2) = HC;
HMC(m+n+1:2*m+n,3:4) = HCs;

tMC = [tC;sC;sC];
sMC = [tC;sC;tC];

[tS,sS,GS,HS,rankdef] = schurc(m,n,tMC,sMC,GMC,HMC,piv);
if rankdef
	error('Rank deficient matrix, algorithm still unsupported.')
end

x = fmatna(b,xi);

GB = zeros(m+n,4+kb);
HB = zeros(m+kb,4+kb);
GB(m+1:m+n,1:4) = GS;
GB(1:m,5:4+kb) = spdiags(sS,0,m,m)*x;
HB(1:m,1:4) = HS;
HB(m+1:m+kb,5:4+kb) = eye(kb);
tB = [sS;tS];
sB = [sS;zeros(kb,1)];
[tx,sx,Gx,Hx] = schurcmv(m,n,tB,sB,GB,HB,0);
et = ones(size(tx));
es = ones(size(sx));
x = spdiags(1./(tx*es'-et*sx.'),0,n,n)*Gx*Hx';

x = fmatna(x,eta,1);

if reale
	x = real(x);
end


function [t,s,G,H,rankdef] = schurc(m,n,t,s,G,H,piv)
%SCHURC computation of the Schur complement of a Cauchy-like matrix.
%   Version optimized for the mosaic matrix MC used in tpls.m.

if nargin < 7
	piv = 1;	% partial pivoting
end
rankdef = 0;

switch piv
case 0
	;
case 1
	;
otherwise
	error('This kind of pivoting is unsupported.')
end

n1 = m + n + n;
n2 = m + n + m;

if exist('tplsmex','file') == 3
	if isreal(t),  t = complex(t);  end
	if isreal(s),  s = complex(s);  end
	if isreal(G),  G = complex(G);  end
	if isreal(H),  H = complex(H);  end
	[t,s,G,H,rankdef] = tplsmex(0,m,n,t,s,G,H,piv);
	t = t(m+n+1:n1);
	s = s(m+n+1:n2);
	G = G(m+n+1:n1,:);
	H = H(m+n+1:n2,:);
	return
else
	warning('MEX file missing: computation may be slow.')
end

l = zeros(n1,1);
u = zeros(1,n2);
p = [1:m+n]';		% rows permutation

v1 = [m+1:m+n]';
d22 = zeros(n,1);		% diagonal elements of block (2,2)
m23 = (G(v1,:)*H(m+n+1,:)') ./ (t(v1)-s(m+n+1));
				% modified elements of block (2,3)
i23 = ones(n,1);		% indexes of modified elements of block (2,3)

for k  =  1:m
	v3 = [k+1:n2]';
	if sum(t(v1) == s(k)), error('Null denominator (1).'), end
	l(v1) = (G(v1,:)*H(k,:)') ./ (t(v1)-s(k));
	l(k) = 1;
	%[lmax imax] = max(abs(l([k;v1])));
	[lmax imax] = max( abs(real(l([k;v1]))) + abs(imag(l([k;v1]))) );
	if piv && (imax > 1)
		imax = imax+m-1;
		if p(imax) == imax
			v2 = [k:imax-1,imax+1:n2]';
			u(imax) = d22(imax-m);
		else
			im = m + n + i23(imax-m);
			v2 = [k:im-1,im+1:n2]';
			u(im) = m23(imax-m);
		end
		if sum(t(imax) == s(v2)), error('Null denominator (2).'), end
		u(v2) = (G(imax,:)*H(v2,:)') ./ (t(imax)-s(v2)).';
		d22(imax-m) = (G(k,:)*H(imax,:)') ./ (t(k)-s(imax));
		m23(imax-m) = 0;
		i23(imax-m) = k;
		l([k imax]) = l([imax k]);
		t([k imax]) = t([imax k]);
		G([k imax],:) = G([imax k],:);
		p([k imax]) = p([imax k]);
	else
		u(k) = 1;
		u([k+1:m,m+n+1:n2]) = 0;
		if sum(t(k) == s(v1)), error('Null denominator (3).'), end
		u(v1) = (G(k,:)*H(v1,:)') ./ (t(k)-s(v1)).';
	end
	l(v1) = l(v1) / l(k);
	d22 = d22 - l(v1) .* u(v1).';		% update diag. el. block (2,2)
	m23 = m23 - l(v1) .* u(m+n+i23).';	% update mod. el. block (2,3)
	u(v3) = u(v3) / l(k);
	G(v1,:) = G(v1,:) - l(v1)*G(k,:);
	H(v3,:) = H(v3,:) - u(v3)'*H(k,:);
end

maxlk = 0;
for k  =  m+1:m+n
	v1 = [k+1:n+k-1]';
	if sum(t(v1) == s(k)), error('Null denominator (4).'), end
	l(v1) = (G(v1,:)*H(k,:)') ./ (t(v1)-s(k));
	l(k) = d22(k-m);
	l(n+k) = 1;
	%[lmax imax] = max(abs(l([k:m+n])));
	[lmax imax] = max( abs(real(l([k:m+n]))) + abs(imag(l([k:m+n]))) );
	if piv && (imax > 1)
		imax = imax+k-1;
		if p(imax) == imax
			v2 = [k:imax-1,imax+1:n2]';
			u(imax) = d22(imax-m);
		else
			im = m + n + i23(imax-m);
			v2 = [k:im-1,im+1:n2]';
			u(im) = m23(imax-m);
		end
		if sum(t(imax) == s(v2)), error('Null denominator (5).'), end
		u(v2) = (G(imax,:)*H(v2,:)') ./ (t(imax)-s(v2)).';
		d22(imax-m) = (G(k,:)*H(imax,:)') ./ (t(k)-s(imax));
		m23(imax-m) = m23(k-m);
		i23(imax-m) = i23(k-m);
		l([k imax]) = l([imax k]);
		t([k imax]) = t([imax k]);
		G([k imax],:) = G([imax k],:);
		p([k imax]) = p([imax k]);
	else
		if p(k) == k
			v2 = [k+1:n2]';
			u(k) = d22(k-m);
		else
			im = m + n + i23(k-m);
			v2 = [k:im-1,im+1:n2]';
			u(im) = m23(k-m);
		end
		if sum(t(k) == s(v2)), error('Null denominator (6).'), end
		u(v2) = (G(k,:)*H(v2,:)') ./ (t(k)-s(v2)).';
	end
	v3 = [k+1:n+k]';
	v4 = [k+1:n2]';
	modlk = abs(real(l(k))) + abs(imag(l(k)));
	if modlk > maxlk,  maxlk = modlk;  end
	if modlk < eps*maxlk
		rankdef = 1;
		return
	end
	l(v3) = l(v3) / l(k);
	v0 = [k+1:m+n]';			% update diag. el. block (2,2)
	d22(v0-m) = d22(v0-m) - l(v0) .* u(v0).';
						% update mod. el. block (2,3)
	m23(v0-m) = m23(v0-m) - l(v0) .* u(m+n+i23(v0-m)).';
	u(v4) = u(v4) / l(k);
	G(v3,:) = G(v3,:) - l(v3)*G(k,:);
	H(v4,:) = H(v4,:) - u(v4)'*H(k,:);
end

t = t(m+n+1:n1);
s = s(m+n+1:n2);
G = G(m+n+1:n1,:);
H = H(m+n+1:n2,:);


function [t,s,G,H] = schurcmv(m,n,t,s,G,H,piv)
%SCHURCMV computation of the product of a vector times a Cauchy-like matrix by
%   means of a Schur complement.
%   Version optimized for the mosaic matrix MC used in tpls.m.

if nargin < 7
	piv = 0;	% no pivoting
end

switch piv
case 0
	;
otherwise
	error('This kind of pivoting is unsupported.')
end

kb = size(G,2) - 4;
n1 = m + n;
n2 = m + kb;

if exist('tplsmex') == 3
	[t,s,G,H,rankdef] = tplsmex(1,m,n,t,s,G,H,piv,kb);
	t = t(m+1:n1);
	s = s(m+1:n2);
	G = G(m+1:n1,:);
	H = H(m+1:n2,:);
	return
else
	warning('MEX file missing: computation may be slow.')
end

l = zeros(n1,1);
u = zeros(1,n2);

v1 = [m+1:m+n]';
v2 = [m+1:m+kb]';
for k  =  1:m
	if sum(t(v1) == s(k)), error('Null denominator (7).'), end
	l(v1) = (G(v1,:)*H(k,:)') ./ (t(v1)-s(k));
%	l(k) = 1;		to be precise, l(k)=-1
	if sum(t(k) == s(v2)), error('Null denominator (8).'), end
	u(v2) = (G(k,:)*H(v2,:)') ./ (t(k)-s(v2)).';
%	u(k) = 1;		unnecessary
%	l(v1) = l(v1) / l(k);	unnecessary if l(k)=1
	l(v1) = -l(v1);		% necessary: l(k)=-1
%	u(v2) = u(v2) / l(k);	unnecessary if l(k)=1
	u(v2) = -u(v2);		% necessary: l(k)=-1
	G(v1,:) = G(v1,:) - l(v1)*G(k,:);
	H(v2,:) = H(v2,:) - u(v2)'*H(k,:);
end 

t = t(m+1:n1);
s = s(m+1:n2);
G = G(m+1:n1,:);
H = H(m+1:n2,:);

